/*
 * Copyright (c) 2012-2014,2016-2019 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *					Anthony Mallet on Wed Mar 14 2012
 */
#include <sys/types.h>
#include <assert.h>
#include <errno.h>
#include <err.h>
#include <fcntl.h>
#ifdef HAVE_FEATURES_H
# include <features.h>
#endif
#include <getopt.h>
#include <pthread.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "csLib.h"
#include "h2devLib.h"
#include "h2initGlob.h"
#include "semLib.h"
#include "taskLib.h"

#include "genom3/c/client.h"
#include "genom-iface.h"
#include "private.h"


/* --- local data ---------------------------------------------------------- */

static int pocolibs_h2init = 0;
static int pocolibs_tdev = ERROR;
static size_t pocolibs_mboxsize = 0;

static pthread_t pocolibs_evtid;
static int pocolibs_evtid_init = 0;
static pthread_mutex_t evmutex = PTHREAD_MUTEX_INITIALIZER;

static genom_client clients = NULL;

static void *	pocolibs_evtask(void *data);


/* --- pocolibs_newclient -------------------------------------------------- */

genom_client
pocolibs_newclient()
{
  genom_client h = malloc(sizeof(*h));
  if (!h) return NULL;

  h->context.raise = genom_pocolibs_client_raise;
  h->context.raised = genom_pocolibs_client_raised;
  h->context.data = &h->context_data;

  h->context_data.h = h;
  h->context_data.ex = genom_ok;
  h->context_data.exdetail = NULL;
  h->context_data.exsize = 0;

  h->instance = NULL;
  h->csid = NULL;
  h->active_rqst = NULL;
  h->nactive = h->nextrqst = 0;

  h->evpending = 0;
  if (pipe(h->evpipe) || fcntl(h->evpipe[0], F_SETFL, O_NONBLOCK)) {
    errnoSet(errno);
    free(h);
    return NULL;
  }

  pthread_mutex_lock(&evmutex);
  h->next = clients;
  clients = h;
  pthread_mutex_unlock(&evmutex);

  return h;
}


/* --- pocolibs_delclient -------------------------------------------------- */

void
pocolibs_delclient(genom_client h)
{
  if (!h) return;

  pthread_mutex_lock(&evmutex);
  if (h == clients) {
    clients = h->next;
  } else {
    genom_client c;

    for(c = clients; c && c->next != h; c = c->next);
    if (c) c->next = h->next;
  }
  pthread_mutex_unlock(&evmutex);

  if (h->csid) csClientEnd(h->csid);
  pocolibs_fini();
  if (h->instance) free(h->instance);
  if (h->active_rqst) free(h->active_rqst);

  close(h->evpipe[0]);
  close(h->evpipe[1]);

  free(h);
}


/* --- pocolibs_init ------------------------------------------------------- */

genom_event
pocolibs_init(genom_client h, size_t mboxsize)
{
  char name[H2_DEV_MAX_NAME];
  STATUS s;
  int e;

  /* general pocolibs init */
  if (!pocolibs_h2init++) {
    s = h2initGlob(0);
    if (s != OK) { pocolibs_h2init--; goto error; }
  }

  /* mailbox init */
  if (pocolibs_mboxsize > 0 && pocolibs_mboxsize < mboxsize) {
    s = csMboxUpdate(0, mboxsize);
    if (s != OK) goto error;
    pocolibs_mboxsize = mboxsize;
  } else if (pocolibs_mboxsize == 0) {
    snprintf(name, sizeof(name), "cclient-%d", getpid());
    s = csMboxInit(name, 0, mboxsize);
    if (s != OK) {
      genom_mwerr_detail d;
      char *p;
      e = errnoGet();
      p = d.what;
      p += snprintf(p, sizeof(d.what), "%s: ", name);
      h2getErrMsg(e, p, sizeof(d.what) - (p - d.what));
      return genom_mwerr(&d, &h->context);
    }
    pocolibs_mboxsize = mboxsize;
    pocolibs_tdev = taskGetUserData(0);
  }

  /* create event task */
  if (!pocolibs_evtid_init) {
    s = pthread_create(&pocolibs_evtid, NULL, pocolibs_evtask, NULL);
    if (s) goto error;
    pocolibs_evtid_init = 1;
  }

  return genom_ok;

error:
  e = errnoGet();
  if (H2_SYS_ERR_FLAG(e))
    return genom_syserr(&(genom_syserr_detail){ .code = e }, &h->context);
  else {
    genom_mwerr_detail d;
    h2getErrMsg(e, d.what, sizeof(d.what));
    return genom_mwerr(&d, &h->context);
  }
  assert(!"not reached");
}


/* --- pocolibs_fini ------------------------------------------------------- */

void
pocolibs_fini(void)
{
  if (pocolibs_h2init) pocolibs_h2init--;
  if (pocolibs_h2init) return;

  if (pocolibs_evtid_init) {
    h2semGive(H2DEV_TASK_SEM_ID(pocolibs_tdev));
    pthread_join(pocolibs_evtid, NULL);
    pocolibs_evtid_init = 0;
  }

  if (pocolibs_mboxsize)
    csMboxEnd();
  pocolibs_mboxsize = 0;
  pocolibs_tdev = ERROR;
}


/* --- pocolibs_getopt ----------------------------------------------------- */

/* options */
static const char shortopts_string[] = "+i:";
static struct option longopts_list[] = {
  { "name",		required_argument,	NULL,	'i'},
  { "maxmsg",		required_argument,	NULL,	-'m'},
  { NULL, 0, NULL, 0}
};

const struct genom_client_option genom_pocolibs_options[] = {
  { "-i|--name",	"instance name" },
  { "--maxmsg",		"maximum message size" },
  { NULL, NULL}
};

genom_event
pocolibs_getopt(genom_client h, int argc, char *argv[],
                struct pocolibs_options *opts)
{
  int c;
  char *e;

  opts->varmsg_maxsize = 32768;
  opts->instance = NULL;

  /* reset getopt */
#ifdef HAVE_OPTRESET
  optreset = 1;
#endif
#ifdef __GLIBC__
  optind = 0; /* not "1", see gnu manual for getopt(3) ... */
#else
  optind = 1;
#endif

  /* parse command line options */
  opterr = 0;
  while (
    (c = getopt_long(argc, argv, shortopts_string, longopts_list, NULL)) != -1)
    switch (c) {
      case 0: break;

      case 'i': opts->instance = optarg; break;
      case -'m': {
        unsigned long s = strtoul(optarg, &e, 0);
        if (*e != 0 || s <= 0)
          return genom_syserr(&(genom_syserr_detail){.code = EINVAL},
                              &h->context);
        opts->varmsg_maxsize = s * 1024;
        break;
      }

      case '?':
      default:
        c = '?'; break;
    }
  if (argc > optind) c = '?';
  if (c == '?')
    return genom_syserr(&(genom_syserr_detail){.code = EINVAL}, &h->context);
  return genom_ok;
}


/* --- pocolibs_newrqst ---------------------------------------------------- */

genom_event
pocolibs_newrqst(genom_client h, size_t outsize, int *rqstid)
{
  struct genom_client_request *a;
  int i;
  assert(h);

  *rqstid = -1;
  for(i = h->nextrqst; i < h->nextrqst + h->nactive; i++)
    if (h->active_rqst[i%h->nactive].status == FREE) {
      *rqstid = i%h->nactive;
      break;
    }
  if (*rqstid < 0) {
    a = realloc(h->active_rqst, (h->nactive+1)*sizeof(*a));
    if (!a)
      return genom_syserr(&(genom_syserr_detail){.code = errno}, &h->context);
    *rqstid = h->nactive;
    h->nactive++;
    h->active_rqst = a;
  }
  h->nextrqst = *rqstid + 1;

  h->evpending = 1;
  h->active_rqst[*rqstid].info = NULL;
  h->active_rqst[*rqstid].rid = -1;
  h->active_rqst[*rqstid].aid = -1;
  h->active_rqst[*rqstid].status = FREE;
  h->active_rqst[*rqstid].csstatus = FREE;
  h->active_rqst[*rqstid].report = genom_ok;
  h->active_rqst[*rqstid].output = malloc(outsize);
  h->active_rqst[*rqstid].exdetail = NULL;
  h->active_rqst[*rqstid].exfini = NULL;
  h->active_rqst[*rqstid].decode = NULL;
  h->active_rqst[*rqstid].sentcb = NULL;
  h->active_rqst[*rqstid].donecb = NULL;
  h->active_rqst[*rqstid].cb_data = NULL;
  if (!h->active_rqst[*rqstid].output)
    return genom_syserr(&(genom_syserr_detail){.code = errno}, &h->context);

  return genom_ok;
}


/* --- pocolibs_actinfo ---------------------------------------------------- */

const struct genom_service_info *
pocolibs_service_info(genom_client h, int rqstid)
{
  if (rqstid < 0 || rqstid >= h->nactive) return NULL;
  if (h->active_rqst[rqstid].status == FREE) return NULL;
  return h->active_rqst[rqstid].info;
}


/* --- pocolibs_waitevent -------------------------------------------------- */

genom_event
pocolibs_waitevent(genom_client h)
{
  fd_set fdset;
  int s;

  FD_ZERO(&fdset);
  do {
    FD_SET(h->evpipe[0], &fdset);
    s = select(h->evpipe[0]+1, &fdset, NULL, NULL, NULL);
  } while(s < 0 && errno == EINTR);

  if (s < 0) return genom_syserr(&(genom_syserr_detail){.code = errno},
                                 &h->context);
  return genom_ok;
}


/* --- pocolibs_waitack ---------------------------------------------------- */

genom_event
pocolibs_waitack(genom_client h, int rqstid)
{
  genom_event s;
  if (rqstid < 0 || rqstid >= h->nactive)
    return genom_syserr(&(genom_syserr_detail){.code = ENOENT}, &h->context);

  while(h->active_rqst[rqstid].status == WAITING_INTERMED_REPLY) {
    s = pocolibs_waitevent(h);
    if (s) return s;

    s = pocolibs_doevents(h);
    if (s) return s;
  }
  return genom_ok;
}


/* --- pocolibs_waitterm --------------------------------------------------- */

genom_event
pocolibs_waitterm(genom_client h, int rqstid)
{
  genom_event s;
  if (rqstid < 0 || rqstid >= h->nactive)
    return genom_syserr(&(genom_syserr_detail){.code = ENOENT}, &h->context);

  while(h->active_rqst[rqstid].status == WAITING_INTERMED_REPLY ||
        h->active_rqst[rqstid].status == WAITING_FINAL_REPLY) {
    s = pocolibs_waitevent(h);
    if (s) return s;

    s = pocolibs_doevents(h);
    if (s) return s;
  }
  return genom_ok;
}


/* --- pocolibs_done ------------------------------------------------------- */

int
pocolibs_done(genom_client h, int rqstid)
{
  if (rqstid < 0 || rqstid >= h->nactive) return 1;
  if (h->active_rqst[rqstid].status == WAITING_INTERMED_REPLY ||
      h->active_rqst[rqstid].status == WAITING_FINAL_REPLY)
    return 0;
  return 1;
}


/* --- pocolibs_result ----------------------------------------------------- */

genom_event
pocolibs_result(genom_client h, int rqstid,
                genom_event *report, void **output, void **exdetail)
{
  struct genom_client_request *a;
  genom_event s;

  s = pocolibs_waitterm(h, rqstid);
  if (s) return s;

  a = &h->active_rqst[rqstid];
  if (a->status == FREE)
    return genom_syserr(&(genom_syserr_detail){.code = ENOENT}, &h->context);

  if (report) *report = a->report;
  if (output) *output = a->output;
  if (exdetail) *exdetail = a->exdetail;
  return genom_ok;
}


/* --- pocolibs_clean ------------------------------------------------------ */

genom_event
pocolibs_clean(genom_client h, int rqstid)
{
  struct genom_client_request *a;
  int e;
  if (rqstid < 0 || rqstid >= h->nactive)
    return genom_syserr(&(genom_syserr_detail){.code = ENOENT}, &h->context);

  a = &h->active_rqst[rqstid];
  if (a->status == FREE)
    return genom_syserr(&(genom_syserr_detail){.code = ENOENT}, &h->context);

  if (a->csstatus != FINAL_REPLY_OK) {
    e = csClientRqstIdFree(h->csid, a->rid);
    if (e) {
      genom_mwerr_detail d;
      h2getErrMsg(errnoGet(), d.what, sizeof(d.what));
      return genom_mwerr(&d, &h->context);
    }
  }

  a->status = FREE;
  if (a->output) {
    if (a->info->fini_output) a->info->fini_output(a->output);
    free(a->output);
  }
  if (a->exdetail) {
    assert(a->report);
    if (a->exfini) a->exfini(a->exdetail);
    free(a->exdetail);
  }
  return genom_ok;
}


/* --- pocolibs_recv ------------------------------------------------------- */

genom_event
pocolibs_recv(genom_client h)
{
  struct genom_client_request *a;
  int i, s;

  /* check all pending requests */
  for(i = 0; i < h->nactive; i++) {
    a = &h->active_rqst[i];
    if (a->status == FREE || a->csstatus == FINAL_REPLY_OK) continue;

    assert(a->exdetail == NULL);

    s = csClientReplyRcv(
      h->csid, a->rid, NO_BLOCK,
      (char *)&a->aid, sizeof(a->aid), NULL, (void *)a, 0, a->decode);
    if (a->csstatus == s) continue;

    a->csstatus = s;
    switch(a->csstatus) {
      case WAITING_INTERMED_REPLY:
      case WAITING_FINAL_REPLY:
      case FINAL_REPLY_OK:
        break;

      default: /* == ERROR */
        a->csstatus = FINAL_REPLY_OK;
        if (a->report == genom_ok) {
          /* middleware error */
          a->report = genom_mwerr_id;
          a->exfini = NULL;
          a->exdetail = malloc(sizeof(genom_mwerr_detail));
          if (a->exdetail) {
            genom_mwerr_detail *d = a->exdetail;
            h2getErrMsg(errnoGet(), d->what, sizeof(d->what));
            d->what[sizeof(d->what)-1] = *"";
          } else
            a->report = genom_fatal;
        }
    }
  }

  return genom_ok;
}


/* --- pocolibs_doevents --------------------------------------------------- */

genom_event
pocolibs_doevents(genom_client h)
{
  struct genom_client_request *a;
  genom_event e;
  int evpending;
  char c[16];
  ssize_t r;
  int i;

  /* flush pipe */
  do {
    do {
      r = read(h->evpipe[0], c, sizeof(c));
    } while (r == -1 && errno == EINTR);
    switch(r) {
      case 0:
        return genom_syserr(&(genom_syserr_detail){.code = EBADF}, &h->context);

      case -1:
        if ((errno == EAGAIN || errno == EWOULDBLOCK)) break;
        return genom_syserr(&(genom_syserr_detail){.code = errno}, &h->context);
    }
  } while(r == sizeof(c));

  /* check all pending requests */
  e = pocolibs_recv(h);
  if (e != genom_ok) return e;

  /* call callbacks for new replies */
  evpending = 0;
  for(i = 0; i < h->nactive; i++) {
    a = &h->active_rqst[i];
    if (a->status != FREE && a->status != FINAL_REPLY_OK) evpending = 1;
    if (a->status == FREE || a->csstatus == FREE) continue;
    if (a->status == a->csstatus) continue;

    a->status = a->csstatus;
    switch(a->csstatus) {
      case WAITING_FINAL_REPLY:
        if (a->sentcb)
          (*a->sentcb)(h, i, 0, genom_ok, NULL, NULL, a->cb_data);
        break;

      case FINAL_REPLY_OK:
        if (a->donecb)
          (*a->donecb)(h, i, 1, a->report, a->output, a->exdetail, a->cb_data);
        break;

      default: assert(!"reached unreachable");
    }
  }
  if (!evpending) h->evpending = 0;

  return genom_ok;
}


/* --- pocolibs_evtask ----------------------------------------------------- */

static void *
pocolibs_evtask(void *data)
{
  genom_client h;
  H2SEM_ID sem;
  ssize_t s;
  char c = 0;

  if (pocolibs_tdev == ERROR) goto done;

  sem = H2DEV_TASK_SEM_ID(pocolibs_tdev);
  while (h2semTake(sem, 0) == TRUE) {
    if (!pocolibs_h2init) break;

    pthread_mutex_lock(&evmutex);
    for(h = clients; h; h = h->next) {
      if (!h->evpending) continue;

      do
        s = write(h->evpipe[1], &c, 1);
      while (s < 0 && errno == EINTR);
    }
    pthread_mutex_unlock(&evmutex);
  }

done:
  pthread_mutex_lock(&evmutex);
  for(h = clients; h; h = h->next) {
    do
      s = write(h->evpipe[1], &c, 1);
    while (s < 0 && errno == EINTR);
  }
  pthread_mutex_unlock(&evmutex);
  return NULL;
}


/* --- exceptions ---------------------------------------------------------- */

genom_event
genom_pocolibs_client_raise(genom_event ex, void *detail, size_t size,
                            genom_context self)
{
  struct genom_context_data *data = self->data;

  if (data->exsize < size) {
    void *exdetail = realloc(data->exdetail, size);
    if (!exdetail) return data->ex = genom_fatal;

    data->exsize = size;
    data->exdetail = exdetail;
  }
  data->ex = ex;
  memcpy(data->exdetail, detail, size);
  return ex;
}

const void *
genom_pocolibs_client_raised(genom_event *ex, genom_context self)
{
  struct genom_context_data *data = self->data;

  if (ex) *ex = data->ex;
  return data->exdetail;
}
